"""Author: Niklas Strømsnes
Date: 2022-01-09
"""

from typing import List
import numpy as np
import pandas as pd
import scipy.signal as signal
import matplotlib.pyplot as plt
from utils.data_processing.preprocessing import window_signals

from utils.global_constants import SAMPLE_RATE, ACTUATOR_1, SENSOR_1, SENSOR_2, SENSOR_3
from utils.objects import Table, Plate, Actuator, Sensor
from utils.little_helpers import distance_between
from utils.data_visualization.drawing import plot_legend_without_duplicates
from utils.data_processing.detect_echoes import find_first_peak_index, get_envelopes


class Setup:
    surface = Plate()
    actuators: np.ndarray
    sensors: np.ndarray

    def __init__(self):
        raise NameError("Setup version needs to be specified")

    def draw(self):
        _, ax = plt.subplots()
        # fig.gcf().set_size_inches(5.5, 3.5)
        self.surface.draw()
        [actuator.draw() for actuator in self.actuators]
        [sensor.draw() for sensor in self.sensors if sensor.plot]
        ax.axis("scaled")
        ax.set_xlabel("x (m)")
        ax.set_ylabel("y (m)")
        plot_legend_without_duplicates()

    def get_propagation_speed(
        self,
        measurements: pd.DataFrame,
        object_1: Sensor or Actuator,
        object_2: Sensor or Actuator,
    ):
        """Use the cross correlation between the two channels
        to find the propagation speed. Based on:
        https://stackoverflow.com/questions/41492882/find-time-shift-of-two-signals-using-cross-correlation
        """
        n = len(measurements[object_1.name])
        corr = signal.correlate(
            measurements[object_1.name], measurements[object_2.name], mode="same"
        )
        delay_arr = np.linspace(
            start=-0.5 * n / SAMPLE_RATE, stop=0.5 * n / SAMPLE_RATE, num=n
        )
        delay = delay_arr[np.argmax(corr)]
        distance = np.linalg.norm(object_1.coordinates - object_2.coordinates)
        propagation_speed = np.abs(distance / delay)
        return propagation_speed


class Setup1(Setup):
    """Sensors in the middle of the table to
    separate direct signal and reflections.
    """

    actuators = np.empty(shape=1, dtype=Actuator)
    sensors = np.empty(shape=3, dtype=Sensor)
    actuators[ACTUATOR_1] = Actuator(
        coordinates=np.array([1 / 2 * Plate.LENGTH - 0.15, 1 / 2 * Plate.WIDTH])
    )
    sensors[SENSOR_1] = Sensor(
        coordinates=(actuators[ACTUATOR_1].coordinates + np.array([0.10, 0])),
        name="Sensor 1",
    )
    sensors[SENSOR_2] = Sensor(
        coordinates=(sensors[SENSOR_1].coordinates + np.array([0.10, 0])),
        name="Sensor 2",
    )
    sensors[SENSOR_3] = Sensor(
        coordinates=(sensors[SENSOR_2].coordinates + np.array([0.10, 0])),
        name="Sensor 3",
    )

    def __init__(self):
        pass

    def get_propagation_speed(
        self,
        measurements: pd.DataFrame,
        prominence: float = 0.001,
        plot_correlation: bool = False,
    ):
        """Use the cross correlation between the two channels
        to find the propagation speed. Based on:
        https://stackoverflow.com/questions/41492882/find-time-shift-of-two-signals-using-cross-correlation
        """
        object_1 = self.sensors[SENSOR_1]
        object_2 = self.sensors[SENSOR_2]
        n = len(measurements[object_1.name])
        corr = signal.correlate(
            measurements[object_1.name], measurements[object_2.name], mode="same"
        )
        delay_arr = np.linspace(
            start=-0.5 * n / SAMPLE_RATE, stop=0.5 * n / SAMPLE_RATE, num=n
        )
        delay = delay_arr[np.argmax(corr)]
        distance = np.linalg.norm(object_1.coordinates - object_2.coordinates)
        propagation_speed_sensors_1_2 = np.abs(distance / delay)
        print(
            f"Propagation speed between sensors 1 and 2: {propagation_speed_sensors_1_2:.2f} m/s"
        )
        # Again, but with sensors 2 and 3
        object_1 = self.sensors[SENSOR_2]
        object_2 = self.sensors[SENSOR_3]
        n = len(measurements[object_1.name])
        corr = signal.correlate(
            measurements[object_1.name], measurements[object_2.name], mode="same"
        )
        delay_arr = np.linspace(
            start=-0.5 * n / SAMPLE_RATE, stop=0.5 * n / SAMPLE_RATE, num=n
        )
        delay = delay_arr[np.argmax(corr)]
        distance = np.linalg.norm(object_1.coordinates - object_2.coordinates)
        propagation_speed_sensors_2_3 = np.abs(distance / delay)
        print(
            f"Propagation speed between sensors 2 and 3: {propagation_speed_sensors_2_3:.2f} m/s"
        )
        # Again, but with sensors 1 and 3q
        object_1 = self.sensors[SENSOR_1]
        object_2 = self.sensors[SENSOR_3]
        n = len(measurements[object_1.name])
        corr = signal.correlate(
            measurements[object_1.name], measurements[object_2.name], mode="same"
        )
        delay_arr = np.linspace(
            start=-0.5 * n / SAMPLE_RATE, stop=0.5 * n / SAMPLE_RATE, num=n
        )
        delay = delay_arr[np.argmax(corr)]
        distance = np.linalg.norm(object_1.coordinates - object_2.coordinates)
        propagation_speed_sensors_1_3 = np.abs(distance / delay)
        print(
            f"Propagation speed between sensors 1 and 3: {propagation_speed_sensors_1_3:.2f} m/s"
        )
        propagation_speed = (
            propagation_speed_sensors_1_2
            + propagation_speed_sensors_2_3
            + propagation_speed_sensors_1_3
        ) / 3
        if plot_correlation:
            fig, ax = plt.subplots()
            ax.plot(delay_arr, corr)
            ax.set_xlabel("Delay (s)")
            ax.set_ylabel("Correlation")
        """Alternatively:"""
        # peak_object1 = np.argmax(np.abs(signal.hilbert(measurements[object_1.name])))
        # peak_object2 = np.argmax(np.abs(signal.hilbert(measurements[object_2.name])))
        # delay = (peak_object2 - peak_object1) / SAMPLE_RATE
        # distance = np.linalg.norm(object_1.coordinates - object_2.coordinates)
        # propagation_speed = np.abs(distance / delay)
        """Or even alternativelier:"""
        # measurements = get_envelopes(measurements)
        # _, ax = plt.subplots()
        # first_peak_object1 = find_first_peak_index(measurements[object_1.name], ax=ax)
        # first_peak_object2 = find_first_peak_index(measurements[object_2.name], ax=ax)
        # delay = (first_peak_object2 - first_peak_object1) / SAMPLE_RATE
        # distance = np.linalg.norm(object_1.coordinates - object_2.coordinates)
        # propagation_speed = np.abs(distance / delay)
        return propagation_speed


class Setup2(Setup):
    """Sensors in a straight line across the full table"""

    actuators = np.empty(shape=1, dtype=Actuator)
    sensors = np.empty(shape=3, dtype=Sensor)

    sensors[SENSOR_1] = Sensor(coordinates=np.array([0.135, 0.305]), name="Sensor 1")
    sensors[SENSOR_2] = Sensor(
        coordinates=(sensors[SENSOR_1].coordinates + np.array([0.267, 0])),
        name="Sensor 2",
    )
    sensors[SENSOR_3] = Sensor(
        sensors[SENSOR_2].coordinates + np.array([0.267, 0]), name="Sensor 3"
    )
    actuators[0] = Actuator(np.array([sensors[SENSOR_1].x / 2, sensors[SENSOR_1].y]))

    def __init__(self):
        pass

    def get_propagation_speed(self, measurements: pd.DataFrame):
        """Use the cross correlation between the two channels
        to find the propagation speed. Based on:
        https://stackoverflow.com/questions/41492882/find-time-shift-of-two-signals-using-cross-correlation
        """
        n = len(measurements[self.sensors[SENSOR_1].name])
        corr = signal.correlate(
            measurements[self.sensors[SENSOR_2].name],
            measurements[self.sensors[SENSOR_1].name],
            mode="same",
        )
        delay_arr = np.linspace(
            start=-0.5 * n / SAMPLE_RATE, stop=0.5 * n / SAMPLE_RATE, num=n
        )
        delay = delay_arr[np.argmax(corr)]
        distance = np.abs(self.sensors[SENSOR_1].x - self.sensors[SENSOR_2].x)
        propagation_speed = np.round(np.abs(distance / delay), decimals=2)
        return propagation_speed


class Setup3(Setup):
    """Inspecting touch signal right under the touch location
    and over and under at"""

    actuators = np.empty(shape=1, dtype=Actuator)
    sensors = np.empty(shape=3, dtype=Sensor)
    actuators[ACTUATOR_1] = Actuator(coordinates=(np.array([0.45, Plate.WIDTH / 2])))
    sensors[SENSOR_1] = Sensor(
        coordinates=actuators[ACTUATOR_1].coordinates, name="Sensor 1"
    )
    sensors[SENSOR_2] = Sensor(
        coordinates=(Plate.LENGTH / 2, Plate.WIDTH / 2), name="Sensor 2"
    )
    sensors[SENSOR_3] = Sensor(
        coordinates=(Plate.LENGTH / 2, Plate.WIDTH / 2), name="Sensor 3"
    )

    def __init__(self):
        pass

    def get_propagation_speed(self, measurements: pd.DataFrame):
        object_1 = self.sensors[SENSOR_1]
        object_2 = self.sensors[SENSOR_3]
        envelopes = get_envelopes(measurements)
        _, ax = plt.subplots()
        first_peak_object1 = find_first_peak_index(envelopes[object_1.name], ax=ax)
        measurements = window_signals(
            measurements,
            length_of_signal_seconds=(0.384 - 0.378),
            peak_index=first_peak_object1,
        )
        n = len(measurements[object_1.name])
        corr = signal.correlate(
            measurements[object_1.name], measurements[object_2.name], mode="same"
        )
        delay_arr = np.linspace(
            start=-0.5 * n / SAMPLE_RATE, stop=0.5 * n / SAMPLE_RATE, num=n
        )
        delay = delay_arr[np.argmax(corr)]
        distance = np.linalg.norm(object_1.coordinates - object_2.coordinates)
        propagation_speed = np.abs(distance / delay)
        return propagation_speed


class Setup_3x3(Setup):
    actuators = np.empty(shape=1, dtype=Actuator)
    sensors = np.empty(shape=9, dtype=Sensor)
    actuators[ACTUATOR_1] = Actuator(
        coordinates=np.array([1 / 2 * Plate.LENGTH - 0.15, 1 / 2 * Plate.WIDTH])
    )
    sensors[1] = Sensor(
        coordinates=np.array([Plate.LENGTH / 2, Plate.WIDTH - 0.05]), name="Sensor 2"
    )
    sensors[0] = Sensor(
        coordinates=(sensors[1].x - 0.02, sensors[1].y), name="Sensor 1"
    )
    sensors[2] = Sensor(
        coordinates=(sensors[1].x + 0.02, sensors[1].y), name="Sensor 3"
    )
    sensors[3] = Sensor(
        coordinates=(sensors[1].x - 0.02, sensors[1].y - 0.02), name="Sensor 4"
    )
    sensors[4] = Sensor(
        coordinates=(sensors[1].x, sensors[1].y - 0.02), name="Sensor 5"
    )
    sensors[5] = Sensor(
        coordinates=(sensors[1].x + 0.02, sensors[1].y - 0.02), name="Sensor 6"
    )
    sensors[6] = Sensor(
        coordinates=(sensors[1].x - 0.02, sensors[1].y - 0.04), name="Sensor 7"
    )
    sensors[7] = Sensor(
        coordinates=(sensors[1].x, sensors[1].y - 0.04), name="Sensor 8"
    )
    sensors[8] = Sensor(
        coordinates=(sensors[1].x + 0.02, sensors[1].y - 0.04), name="Sensor 9"
    )

    def __init__(self):
        pass

    def get_propagation_speed(
        self, measurements: pd.DataFrame, prominence: float = 0.001
    ):
        """Use the cross correlation between the two channels
        to find the propagation speed. Based on:
        https://stackoverflow.com/questions/41492882/find-time-shift-of-two-signals-using-cross-correlation
        """
        object_1 = self.sensors[SENSOR_1]
        object_2 = self.sensors[SENSOR_3]
        # n = len(measurements[object_1.name])
        # corr = signal.correlate(measurements[object_1.name],
        #                         measurements[object_2.name],
        #                         mode='same')
        # delay_arr = np.linspace(start=-0.5 * n / SAMPLE_RATE,
        #                         stop=0.5 * n / SAMPLE_RATE,
        #                         num=n)
        # delay = delay_arr[np.argmax(corr)]
        # distance = np.linalg.norm(object_1.coordinates - object_2.coordinates)
        # propagation_speed = np.abs(distance / delay)
        """Alternatively:"""
        # peak_object1 = np.argmax(np.abs(signal.hilbert(measurements[object_1.name])))
        # peak_object2 = np.argmax(np.abs(signal.hilbert(measurements[object_2.name])))
        # delay = (peak_object2 - peak_object1) / SAMPLE_RATE
        # distance = np.linalg.norm(object_1.coordinates - object_2.coordinates)
        # propagation_speed = np.abs(distance / delay)
        """Or even alternativelier:"""
        measurements = get_envelopes(measurements)
        _, ax = plt.subplots()
        first_peak_object1 = find_first_peak_index(measurements[object_1.name], ax=ax)
        first_peak_object2 = find_first_peak_index(measurements[object_2.name], ax=ax)
        delay = (first_peak_object2 - first_peak_object1) / SAMPLE_RATE
        distance = np.linalg.norm(object_1.coordinates - object_2.coordinates)
        propagation_speed = np.abs(distance / delay)
        return propagation_speed


class Setup_ULA(Setup):
    """A line of number_of_sensors sensors,
    with an actuator given by actuator_coordinates.
    """

    def __init__(
        self,
        number_of_sensors: int,
        actuator_coordinates: np.ndarray or List[np.ndarray],
        array_start_coordinates: np.ndarray,
        array_spacing_m: float,
    ):
        self.number_of_sensors = number_of_sensors
        self.array_start_coordinates = array_start_coordinates
        self.array_spacing = array_spacing_m
        self.sensors = np.empty(shape=number_of_sensors, dtype=Sensor)
        for i in range(number_of_sensors):
            self.sensors[i] = Sensor(
                coordinates=np.array(
                    [
                        array_start_coordinates[0],
                        array_start_coordinates[1] + i * array_spacing_m,
                    ]
                ),
                name=f"Sensor {i + 1}",
            )
        if isinstance(actuator_coordinates, List):
            self.actuators = np.empty(shape=len(actuator_coordinates), dtype=Actuator)
            for i, actuator in enumerate(self.actuators):
                self.actuators[i] = Actuator(coordinates=actuator_coordinates[i])
        else:
            self.actuators = np.empty(shape=1, dtype=Actuator)
            self.actuators[0] = Actuator(coordinates=actuator_coordinates)


class Setup4(Setup_ULA):
    # Setup_Linear_Array with arugments for a 3 sensor array
    def __init__(self, actuator_coordinates: np.ndarray):
        super().__init__(
            number_of_sensors=3,
            actuator_coordinates=actuator_coordinates,
            array_start_coordinates=np.array([0.45, 0.65]),
            array_spacing_m=0.01,
        )

    def get_propagation_speed(
        self, measurements: pd.DataFrame, prominence: float = 0.001
    ):
        """Use the cross correlation between the two channels
        to find the propagation speed.
        """
        object1 = self.sensors[SENSOR_1]
        object2 = self.sensors[SENSOR_2]
        # noise_max_value_object1 = get_noise_max_value(measurements[object1.name],
        #                                               time_window_s=0.1,)
        # noise_max_value_object2 = get_noise_max_value(measurements[object2.name],
        #                                               time_window_s=0.1,)
        # first_rise_object1 = (
        #     measurements[object1.name] > (2 * noise_max_value_object1)).idxmax()
        # first_rise_object2 = (
        #     measurements[object2.name] > (2 * noise_max_value_object2)).idxmax()
        # first_rise_object1 = (
        #     measurements[object1.name] > (np.max(measurements[object1.name]) / 2)).idxmax()
        # first_rise_object2 = (
        #     measurements[object2.name] > (np.max(measurements[object2.name]) / 2)).idxmax()
        # delay = (first_rise_object2 - first_rise_object1) / SAMPLE_RATE
        # assert delay != 0, f"""Delay is 0: First rise indices are {first_rise_object1} and {first_rise_object2}"""
        distance = np.linalg.norm(object1.coordinates - object2.coordinates)
        # propagation_speed = np.abs(distance / delay)
        # Or by max value:
        peak_object1 = np.argmax(np.abs(signal.hilbert(measurements[object1.name])))
        peak_object2 = np.argmax(np.abs(signal.hilbert(measurements[object2.name])))
        delay = (peak_object2 - peak_object1) / SAMPLE_RATE
        souce_distance_cat = np.abs(self.actuators[0].x - object1.x)
        source_distance_hyp = distance_between(
            self.actuators[0].coordinates, object1.coordinates
        )
        angle = np.arccos((souce_distance_cat / source_distance_hyp))
        propagation_speed = np.abs(distance / (delay / np.cos(angle)))
        return propagation_speed


class Setup5(Setup_ULA):
    # Setup_Linear_Array with arugments for a 3 sensor array
    def __init__(
        self,
        actuator_coordinates: np.ndarray,
        array_start_coordinates: np.ndarray = np.array([0.05, 0.05]),
        number_of_sensors: int = 8,
        array_spacing_m: float = 0.01,
    ):
        super().__init__(
            number_of_sensors=number_of_sensors,
            actuator_coordinates=actuator_coordinates,
            array_start_coordinates=array_start_coordinates,
            array_spacing_m=array_spacing_m,
        )


class Setup_UCA(Setup):
    """A line of number_of_sensors sensors,
    with an actuator given by actuator_coordinates.
    """

    def __init__(
        self,
        number_of_sensors: int,
        actuator_coordinates: np.ndarray or List[np.ndarray],
        array_center_coordinates: np.ndarray,
        array_spacing_m: float,
    ):
        self.number_of_sensors = number_of_sensors
        self.array_start_coordinates = array_center_coordinates
        self.array_spacing = array_spacing_m
        self.actuators = np.empty(shape=len(actuator_coordinates), dtype=Actuator)
        self.sensors = np.empty(shape=number_of_sensors, dtype=Sensor)
        angles_of_sensors = np.linspace(0, 2 * np.pi, number_of_sensors, endpoint=False)
        angle_from_sensor_to_sensor = (np.pi - angles_of_sensors[1]) / 2
        radius = (
            array_spacing_m
            * np.sin(angle_from_sensor_to_sensor)
            / np.sin(angles_of_sensors[1])
        )
        for i in range(number_of_sensors):
            self.sensors[i] = Sensor(
                coordinates=np.array(
                    [
                        array_center_coordinates[0]
                        + radius * np.cos(angles_of_sensors[i]),
                        array_center_coordinates[1]
                        + radius * np.sin(angles_of_sensors[i]),
                    ]
                ),
                name=f"Sensor {i + 1}",
            )
        if isinstance(actuator_coordinates, List):
            self.actuators = np.empty(shape=len(actuator_coordinates), dtype=Actuator)
            for i, actuator in enumerate(self.actuators):
                self.actuators[i] = Actuator(coordinates=actuator_coordinates[i])
        else:
            self.actuators = np.empty(shape=1, dtype=Actuator)
            self.actuators[0] = Actuator(coordinates=actuator_coordinates)


class Setup6(Setup_UCA):
    def __init__(
        self,
        actuator_coordinates: np.ndarray,
        array_center_coordinates: np.ndarray = np.array([0.05, 0.05]),
        number_of_sensors: int = 8,
        array_spacing_m: float = 0.01,
    ):
        super().__init__(
            number_of_sensors=number_of_sensors,
            actuator_coordinates=actuator_coordinates,
            array_center_coordinates=array_center_coordinates,
            array_spacing_m=array_spacing_m,
        )
